import numpy as np
import strax
from straxen.plugins.peaks.peak_basics import PeakBasics

export, __all__ = strax.exporter()


@export
class PeakBasicsSOM(PeakBasics):
    """Adds SOM fields to peak basics to be propgated to event basics."""

    __version__ = "0.0.1"
    child_plugin = True

    def infer_dtype(self):
        dtype = super().infer_dtype()
        additional_fields = [
            ("som_sub_type", np.int32, "SOM subtype of the peak(let)"),
            ("straxen_type", np.int8, "Old straxen type of the peak(let)"),
            ("loc_x_som", np.int16, "x location of the peak(let) in the SOM"),
            ("loc_y_som", np.int16, "y location of the peak(let) in the SOM"),
        ]

        return dtype + additional_fields

    def compute(self, peaks):
        peak_basics = super().compute(peaks)
        fields_to_copy = ("som_sub_type", "straxen_type", "loc_x_som", "loc_y_som")
        strax.copy_to_buffer(peaks, peak_basics, "_copy_som_information", fields_to_copy)
        return peak_basics



import sys
from sys import getsizeof, stderr
import inspect
import warnings
import datetime
import pytz
from itertools import chain
from collections import defaultdict, OrderedDict, deque
from importlib import import_module
from platform import python_version
import typing as ty

import numpy as np
import pandas as pd
import socket
import strax
import straxen
from git import Repo, InvalidGitRepositoryError
from configparser import NoSectionError

try:
    # pylint: disable=redefined-builtin
    from reprlib import repr
except ImportError:
    pass

_is_jupyter = any("jupyter" in arg for arg in sys.argv)


export, __all__ = strax.exporter()


@export
def dataframe_to_wiki(df, float_digits=5, title="Awesome table", force_int: ty.Tuple = ()):
    """Convert a pandas dataframe to a dokuwiki table (which you can copy-paste onto the XENON wiki)

    :param df: dataframe to convert
    :param float_digits: format float to this number of digits.
    :param title: title of the table.
    :param force_int: tuple of column names to force to be integers

    """
    table = "^ %s " % title + "^" * (len(df.columns) - 1) + "^\n"
    table += "^ " + " ^ ".join(df.columns) + " ^\n"

    def format_float(x):
        if isinstance(x, float):
            return f"{x:.{float_digits}f}"
        return x

    force_int = np.where(np.in1d(df.columns.values, strax.to_str_tuple(force_int)))[0]

    for _, row in df.iterrows():
        table += (
            "| "
            + " | ".join([
                str(int(x) if i in force_int else format_float(x))
                for i, x in enumerate(row.values.tolist())
            ])
            + " |\n"
        )
    return table


@export
def print_versions(
    modules=("strax", "straxen", "cutax"),
    print_output=not _is_jupyter,
    include_python=True,
    return_string=False,
    include_git=True,
):
    """Print versions of modules installed.

    :param modules: Modules to print, should be str, tuple or list. E.g.
        print_versions(modules=('numpy', 'dddm',))
    :param return_string: optional. Instead of printing the message, return a string
    :param include_git: Include the current branch and latest commit hash
    :return: optional, the message that would have been printed

    """
    versions = defaultdict(list)
    if include_python:
        versions["module"] = ["python"]
        versions["version"] = [python_version()]
        versions["path"] = [sys.executable]
        versions["git"] = [None]
    for m in strax.to_str_tuple(modules):
        result = _version_info_for_module(m, include_git=include_git)
        if result is None:
            continue
        version, path, git_info = result
        versions["module"].append(m)
        versions["version"].append(version)
        versions["path"].append(path)
        versions["git"].append(git_info)
    df = pd.DataFrame(versions)
    info = f"Host {socket.getfqdn()}\n{df.to_string(index=False,)}"
    if print_output:
        print(info)
    if return_string:
        return info
    return df


def _version_info_for_module(module_name, include_git):
    try:
        mod = import_module(module_name)
    except (ModuleNotFoundError, ImportError):
        print(f"{module_name} is not installed")
        return
    git = None
    version = mod.__dict__.get("__version__", None)
    module_path = mod.__dict__.get("__path__", [None])[0]
    if include_git:
        try:
            repo = Repo(module_path, search_parent_directories=True)
        except InvalidGitRepositoryError:
            # not a git repo
            pass
        else:
            try:
                branch = repo.active_branch
            except TypeError:
                branch = "unknown"
            try:
                commit_hash = repo.head.object.hexsha
            except TypeError:
                commit_hash = "unknown"
            git = f"branch:{branch} | {commit_hash[:7]}"
    return version, module_path, git


@export
def utilix_is_configured(
    header: str = "RunDB",
    section: str = "xent_database",
    warning_message: ty.Union[None, bool, str] = None,
) -> bool:
    """Check if we have the right connection to.

    :param header: Which header to check in the utilix config file
    :param section: Which entry in the header to check to exist
    :param warning_message: If utilix is not configured, warn the user. if None -> generic warning
        if str -> use the string to warn if False -> don't warn
    :return: bool, can we connect to the Mongo database?

    """
    try:
        is_configured = (
            hasattr(straxen.uconfig, "get") and straxen.uconfig.get(header, section) is not None
        )
    except NoSectionError:
        is_configured = False

    should_report = bool(warning_message) or warning_message is None
    if not is_configured and should_report:
        if warning_message is None:
            warning_message = "Utilix is not configured, cannot proceed"
        warnings.warn(str(warning_message))
    return is_configured


@export
class TimeWidgets:
    def __init__(self):
        """Creates interactive time widgets which allow to convert a human readble date and time
        into a start and endtime in unix time nano- seconds."""
        self._created_widgets = False

    def create_widgets(self):
        """Creates time and time zone widget for simpler time querying.

        Note:
            Please be aware that the correct format for the time field
            is HH:MM.

        """
        utcend = datetime.datetime.utcnow()
        deltat = datetime.timedelta(minutes=60)
        utcstart = utcend - deltat
        import ipywidgets as widgets

        self._start_widget = self._create_date_and_time_widget(utcstart, "Start")
        self._end_widget = self._create_date_and_time_widget(utcend, "End")
        self._time_zone_widget = self._create_time_zone_widget()
        self._created_widgets = True

        return widgets.VBox([
            widgets.HBox([self._time_zone_widget, widgets.HTML(value="ns:")]),
            self._start_widget,
            self._end_widget,
        ])

    def get_start_end(self):
        """Returns start and end time of the specfied time interval in nano- seconds utc unix
        time."""
        if not self._created_widgets:
            raise ValueError('Please run first "create_widgets". ')

        time_zone = self._time_zone_widget.options[self._time_zone_widget.value][0]

        start, start_ns = self._convert_to_datetime(self._start_widget, time_zone)
        start = start.astimezone(tz=pytz.UTC)
        start = start.timestamp()
        start = int(start * 10**9) + start_ns
        end, end_ns = self._convert_to_datetime(self._end_widget, time_zone)
        end = end.astimezone(tz=pytz.UTC)
        end = end.timestamp()
        end = int(end * 10**9) + end_ns

        if start > end:
            warnings.warn("Start time is larger than endtime are you sure you wanted this?")
        return start, end

    @staticmethod
    def _create_date_and_time_widget(date_and_time, widget_describtion):
        import ipywidgets as widgets
        from ipywidgets import Layout

        date = datetime.date(date_and_time.year, date_and_time.month, date_and_time.day)
        date = widgets.DatePicker(
            description=widget_describtion, value=date, layout=Layout(width="225px"), disabled=False
        )

        time = "{:02d}:{:02d}".format(int(date_and_time.hour), int(date_and_time.minute))
        time = widgets.Text(value=time, layout=Layout(width="75px"), disabled=False)

        time_ns = widgets.Text(value="0", layout=Layout(width="150px"), disabled=False)

        return widgets.HBox([date, time, time_ns])

    @staticmethod
    def _create_time_zone_widget():
        import ipywidgets as widgets

        _time_zone_widget = widgets.Dropdown(
            options=[("CET", 0), ("UTC", 1)],
            value=0,
            description="Time Zone:",
        )
        return _time_zone_widget

    @staticmethod
    def _convert_to_datetime(time_widget, time_zone):
        """Converts values of widget into a timezone aware datetime object.

        :param time_widget: Widget Box containing a DatePicker and two text widget. The first text
            widget is used to set a day time. The second the time in nano-seconds.
        :param time_zone: pytz.timezone allowed string for a timezone.
        :return: timezone aware datetime object.

        """
        date_and_time = [c.value for c in time_widget.children]
        try:
            hour_and_minutes = datetime.datetime.strptime(date_and_time[1], "%H:%M")
        except ValueError as e:
            raise ValueError(
                "Cannot convert time into datetime object. "
                f"Expected the following formating HH:MM. {e}"
            )

        time = datetime.datetime.combine(date_and_time[0], datetime.time())
        time = time.replace(hour=hour_and_minutes.hour, minute=hour_and_minutes.minute)
        time_zone = pytz.timezone(time_zone)
        time = time_zone.localize(time)

        time_ns = int(date_and_time[2])

        return time, time_ns


@strax.Context.add_method
def extract_latest_comment(self):
    """Extract the latest comment in the runs-database. This just adds info to st.runs.

    Example:
        st.extract_latest_comment()
        st.select_runs(available=('raw_records'))

    """
    if self.runs is None or "comments" not in self.runs.keys():
        self.scan_runs(store_fields=("comments",))
        latest_comments = _parse_to_last_comment(self.runs["comments"])
        self.runs["comments"] = latest_comments
    return self.runs


def _parse_to_last_comment(comments):
    """Unpack to get the last comment (hence the -1) or give '' when there is none."""
    return [(c[-1]["comment"] if hasattr(c, "__len__") else "") for c in comments]


@export
def convert_array_to_df(array: np.ndarray) -> pd.DataFrame:
    """Converts the specified array into a DataFrame drops all higher dimensional fields during the
    process.

    :param array: numpy.array to be converted.
    :return: DataFrame with higher dimensions dropped.

    """
    keys = [key for key in array.dtype.names if array[key].ndim == 1]
    return pd.DataFrame(array[keys])


@export
def filter_kwargs(func, kwargs):
    """Filter out keyword arguments that are not in the call signature of func and return filtered
    kwargs dictionary."""
    params = inspect.signature(func).parameters
    if any([str(p).startswith("**") for p in params.values()]):
        # if func accepts wildcard kwargs, return all
        return kwargs
    return {k: v for k, v in kwargs.items() if k in params}


@export
class CacheDict(OrderedDict):
    """Dict with a limited length, ejecting LRUs as needed.

    copied from
    https://gist.github.com/davesteele/44793cd0348f59f8fadd49d7799bd306

    """

    def __init__(self, *args, cache_len: int = 10, **kwargs):
        assert cache_len > 0
        self.cache_len = cache_len

        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        super().move_to_end(key)

        while len(self) > self.cache_len:
            oldkey = next(iter(self))
            super().__delitem__(oldkey)

    def __getitem__(self, key):
        val = super().__getitem__(key)
        super().move_to_end(key)

        return val


@export
def total_size(o, handlers=None, verbose=False):
    """Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    from: https://code.activestate.com/recipes/577504/

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    if handlers is not None:
        all_handlers.update(handlers)  # user handlers take precedence
    seen = set()  # track which object id's have already been seen
    default_size = getsizeof(0)  # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:  # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)



import strax
import numpy as np


class EventwBayesClass(strax.Plugin):
    """Append at event level the posterior probability for an S1, S2, alt_S1 and alt_S2."""

    provides = "event_w_bayes_class"
    depends_on = ("peak_classification_bayes", "event_basics")
    data_kind = "events"
    __version__ = "0.0.1"

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields
        for name in ["s1", "s2", "alt_s1", "alt_s2"]:
            dtype += [(f"{name}_ln_prob_s1", np.float32, f"Given an {name}, s1 ln probability")]
            dtype += [(f"{name}_ln_prob_s2", np.float32, f"Given an {name}, s2 ln probability")]

        return dtype

    def compute(self, peaks, events):
        result = np.empty(len(events), dtype=self.dtype)

        for name in ["s1", "s2", "alt_s1", "alt_s2"]:
            result[f"{name}_ln_prob_s1"] = np.nan
            result[f"{name}_ln_prob_s2"] = np.nan
            # Select peaks based on their start time
            mask = np.in1d(peaks["time"], events[f"{name}_time"])
            mask_ev = np.in1d(events[f"{name}_time"], peaks["time"])
            result[f"{name}_ln_prob_s1"][mask_ev] = peaks["ln_prob_s1"][mask]
            result[f"{name}_ln_prob_s2"][mask_ev] = peaks["ln_prob_s2"][mask]
            result["time"] = events["time"]
            result["endtime"] = events["endtime"]

        return result



import strax
import numpy as np

from straxen.plugins.events.event_basics import EventBasics

export, __all__ = strax.exporter()


@export
class EventBasicsSOM(EventBasics):
    """Adds SOM fields for S1 and S2 peaks to event basics."""

    __version__ = "0.0.1"
    child_plugin = True

    def _set_dtype_requirements(self):
        # Properties to store for each peak (main and alternate S1 and S2)
        # Add here SOM types:
        super()._set_dtype_requirements()
        self.peak_properties = list(self.peak_properties)
        self.peak_properties += [
            ("som_sub_type", np.int32, "SOM subtype of the peak(let)"),
            ("straxen_type", np.int8, "Old straxen type of the peak(let)"),
            ("loc_x_som", np.int16, "x location of the peak(let) in the SOM"),
            ("loc_y_som", np.int16, "y location of the peak(let) in the SOM"),
        ]
        self.peak_properties = tuple(self.peak_properties)

    def compute(self, events, peaks):
        result = super().compute(events, peaks)
        return result


@export
class EventSOMClassification(strax.Plugin):
    """Plugin which propagates S1 SOM infromation to events as long straxen.PeakletClassification is
    still used."""

    depends_on = ("event_basics", "peak_som_classification")
    __version__ = "0.0.1"

    provides = "event_som_classification"

    def infer_dtype(self):
        dtype_peaklets = strax.time_fields + [
            ("s1_som_sub_type", np.int32, "Main S1 SOM subtype of the peak(let)"),
            ("s1_som_type", np.int8, "Main S1 SOM type of the peak(let)"),
            ("s1_loc_x_som", np.int16, "Main S1 x location of the peak(let) in the SOM"),
            ("s1_loc_y_som", np.int16, "Main S1 y location of the peak(let) in the SOM"),
            ("alt_s1_som_sub_type", np.int32, "Alt S1 SOM subtype of the peak(let)"),
            ("alt_s1_som_type", np.int8, "Alt S1 SOM type of the peak(let)"),
            ("alt_s1_loc_x_som", np.int16, "Alt S1 x location of the peak(let) in the SOM"),
            ("alt_s1_loc_y_som", np.int16, "Alt S1 y location of the peak(let) in the SOM"),
        ]
        return dtype_peaklets

    def propagate_field_to_event(self, peaks, events, events_som):
        peaks_in_event = strax.split_by_containment(peaks, events)
        return _propagate_field_to_event(peaks_in_event, events, events_som)

    def compute(self, peaks, events):
        events_som = np.zeros(len(events), self.dtype)
        events_som[:] = -1
        events_som["time"] = events["time"]
        events_som["endtime"] = events["endtime"]

        events_som = self.propagate_field_to_event(peaks, events, events_som)

        return events_som


def _propagate_field_to_event(peaks_in_event, events, events_som):

    for p_in_e, event, event_som in zip(peaks_in_event, events, events_som):
        main_s1 = p_in_e[event["s1_index"]]
        _main_s1_exist = event["s1_index"] != -1
        if _main_s1_exist:
            assert main_s1["time"] == event["s1_time"]
            event_som["s1_som_sub_type"] = main_s1["som_sub_type"]
            event_som["s1_som_type"] = main_s1["som_type"]
            event_som["s1_loc_x_som"] = main_s1["loc_x_som"]
            event_som["s1_loc_y_som"] = main_s1["loc_y_som"]

        _alt_s1_exist = event["alt_s1_index"] != -1
        if _alt_s1_exist:
            alt_s1 = p_in_e[event["alt_s1_index"]]
            assert alt_s1["time"] == event["alt_s1_time"], (alt_s1["time"], event["alt_s1_time"])
            event_som["alt_s1_som_sub_type"] = alt_s1["som_sub_type"]
            event_som["alt_s1_som_type"] = alt_s1["som_type"]
            event_som["alt_s1_loc_x_som"] = alt_s1["loc_x_som"]
            event_som["alt_s1_loc_y_som"] = alt_s1["loc_y_som"]

    return events_som



from hypothesis import given, settings
import strax.testutils
import straxen
import numpy as np


@settings(deadline=None, max_examples=10)
@given(strax.testutils.several_fake_records)
def test_count_pulses(records):
    _check_pulse_count(records)


@settings(deadline=None, max_examples=10)
@given(strax.testutils.several_fake_records_one_channel)
def test_count_pulses_2(records):
    _check_pulse_count(records)


def _check_pulse_count(records):
    # numba starts complaining if n_channels == 1, maybe file bug?
    n_ch = records["channel"].max() + 2 if len(records) else 0
    counts = straxen.plugins.records.count_pulses(records, n_channels=n_ch)

    assert counts.dtype == straxen.pulse_count_dtype(n_ch)

    if not len(records):
        assert len(counts) == 0
        return

    assert len(counts) == 1
    count = counts[0]

    # Check total pulse count and area
    for ch, n in enumerate(counts[0]["pulse_count"]):
        rc = records[records["channel"] == ch]
        rc0 = rc[rc["record_i"] == 0]

        assert n == len(rc0)
        assert count["pulse_area"][ch] == rc["area"].sum()

        # Not sure how to check lone pulses other than duplicating logic
        # already in count_pulses, so just do a basic check:
        assert count["lone_pulse_area"][ch] <= count["pulse_area"][ch]

        # Check baseline values:
        # nan does not exist for integer:
        if len(rc0):
            mean = (
                straxen.NO_PULSE_COUNTS
                if np.isnan(np.mean(rc0["baseline"]))
                else int(np.mean(rc0["baseline"]))
            )
            assert count["baseline_mean"][ch] == mean
            if not np.isnan(count["baseline_rms_mean"][ch]):
                assert count["baseline_rms_mean"][ch] == np.mean(rc0["baseline_rms"])
            else:
                assert np.isnan(np.mean(rc0["baseline_rms"]))



import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@straxen.mini_analysis(requires=("event_basics",))
def load_corrected_positions(
    context,
    run_id,
    events,
    alt_s1=False,
    alt_s2=False,
    cmt_version=None,
    posrec_algos=("mlp", "gcn", "cnn"),
):
    """Returns the corrected position for each position algorithm available, without the need to
    reprocess event_basics, as the needed information is already stored in event_basics.

    :param alt_s1: False by default, if True it uses alternative S1 as main one
    :param alt_s2: False by default, if True it uses alternative S2 as main one
    :param cmt_version: CMT version to use (it can be a list of same length as posrec_algos, if
        different versions are required for different posrec algorithms, default 'local_ONLINE')
    :param posrec_algos: list of position reconstruction algorithms to use (default ['mlp', 'gcn',
        'cnn'])

    """

    posrec_algos = strax.to_str_tuple(posrec_algos)

    if cmt_version is None:
        fdc_config = context.get_single_plugin(run_id, "event_positions").config["fdc_map"]
        if isinstance(fdc_config, str) and "cmt://" in fdc_config:
            cmt_version = straxen.URLConfig.split_url_kwargs(fdc_config)[1].get("version", "ONLINE")
        elif straxen.is_cmt_option(fdc_config):
            cmt_version = fdc_config[1]
        else:
            raise ValueError("FDC map is not a CMT option, cannot infer cmt version.")

    if isinstance(cmt_version, (tuple, list)) and len(cmt_version) != len(posrec_algos):
        raise TypeError(
            f"cmt_version is a list but does not match the posrec_algos ({posrec_algos}) length."
        )

    cmt_version = (
        (cmt_version,) * len(posrec_algos) if isinstance(cmt_version, str) else cmt_version
    )

    # Get drift from CMT
    ep = context.get_single_plugin(run_id, "event_positions")
    drift_speed = ep.electron_drift_velocity
    drift_time_gate = ep.electron_drift_time_gate

    dtype = load_dtypes(posrec_algos)
    result = np.zeros(len(events), dtype=dtype)

    s1_pre = "alt_" if alt_s1 else ""
    s2_pre = "alt_" if alt_s2 else ""
    drift_time = (
        events["drift_time"]
        if not (alt_s1 or alt_s2)
        else events[s2_pre + "s2_center_time"] - events[s1_pre + "s1_center_time"]
    )

    z_obs = -drift_speed * drift_time

    for algo, v_cmt in zip(posrec_algos, cmt_version):
        fdc_tmp = (f"fdc_map_{algo}", v_cmt, True)
        map_tmp = straxen.get_correction_from_cmt(run_id, fdc_tmp)
        itp_tmp = straxen.InterpolatingMap(straxen.common.get_resource(map_tmp, fmt="binary"))
        itp_tmp.scale_coordinates([1.0, 1.0, -drift_speed])

        orig_pos = np.vstack(
            [events[f"{s2_pre}s2_x_{algo}"], events[f"{s2_pre}s2_y_{algo}"], z_obs]
        ).T
        r_obs = np.linalg.norm(orig_pos[:, :2], axis=1)
        delta_r = itp_tmp(orig_pos)
        z_obs = z_obs + drift_speed * drift_time_gate

        # apply radial correction
        with np.errstate(invalid="ignore", divide="ignore"):
            r_cor = r_obs + delta_r
            scale = r_cor / r_obs

        with np.errstate(invalid="ignore"):
            z_cor = -((z_obs**2 - delta_r**2) ** 0.5)
            invalid = np.abs(z_obs) < np.abs(delta_r)
        z_cor[invalid] = z_obs[invalid]

        result[f"x_{algo}"] = orig_pos[:, 0] * scale
        result[f"y_{algo}"] = orig_pos[:, 1] * scale
        result[f"r_{algo}"] = r_cor
        result[f"r_naive_{algo}"] = r_obs
        result[f"r_field_distortion_correction_{algo}"] = delta_r
        result[f"theta_{algo}"] = np.arctan2(orig_pos[:, 1], orig_pos[:, 0])
        result[f"z_{algo}"] = z_cor

    result["z_naive"] = z_obs
    return result


def load_dtypes(posrec_algos):
    dtype = []

    for algo in posrec_algos:
        for xyzr in "x y z r".split():
            dtype += [
                (
                    (
                        (
                            f"Interaction {xyzr}-position, field-distortion corrected (cm) - "
                            f"{algo.upper()} posrec algorithm"
                        ),
                        f"{xyzr}_{algo}",
                    ),
                    np.float32,
                ),
            ]
        dtype += [
            (
                (
                    (
                        "Interaction r-position using observed S2 positions directly (cm) -"
                        f" {algo.upper()} posrec algorithm"
                    ),
                    f"r_naive_{algo}",
                ),
                np.float32,
            ),
            (
                (
                    (
                        "Correction added to r_naive for field distortion (cm) - "
                        f"{algo.upper()} posrec algorithm"
                    ),
                    f"r_field_distortion_correction_{algo}",
                ),
                np.float32,
            ),
            (
                (
                    f"Interaction angular position (radians) - {algo.upper()} posrec algorithm",
                    f"theta_{algo}",
                ),
                np.float32,
            ),
        ]

    dtype += [
        (("Interaction z-position using mean drift velocity only (cm)", "z_naive"), np.float32)
    ]
    return dtype



"""For all of the context, do a quick check to see that we are able to search a field (i.e. can
build the dependencies in the context correctly) See issue #233 and PR #236."""

import os
import unittest
import tempfile
import straxen
from straxen.contexts import xenon1t_dali, xenon1t_led, fake_daq, demo
from straxen.contexts import xenonnt_led, xenonnt_online, xenonnt


##
# XENONnT
##


def test_xenonnt_online():
    st = xenonnt_online(_database_init=False)
    st.search_field("time")


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_xenonnt_online_with_online_frontend():
    st = xenonnt_online(include_online_monitor=True)
    for sf in st.storage:
        if "OnlineMonitor" == sf.__class__.__name__:
            break
    else:
        raise ValueError(f"Online monitor not in {st.storage}")


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_xenonnt_online_rucio_local():
    st = xenonnt_online(include_rucio_local=True, _rucio_local_path="./test")
    for sf in st.storage:
        if "RucioLocalFrontend" == sf.__class__.__name__:
            break
    else:
        raise ValueError(f"Online monitor not in {st.storage}")


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_xennonnt():
    st = xenonnt(_database_init=False)
    st.search_field("time")


def test_xenonnt_led():
    st = xenonnt_led(_database_init=False)
    st.search_field("time")


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_nt_is_nt_online():
    # Test that nT and nT online are the same
    st_online = xenonnt_online(_database_init=False)

    st = xenonnt(_database_init=False)
    for plugin in st._plugin_class_registry.keys():
        print(f"Checking {plugin}")
        nt_key = st.key_for("0", plugin)
        nt_online_key = st_online.key_for("0", plugin)
        assert str(nt_key) == str(nt_online_key)


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_cmt_versions():
    """Let's try and see which CMT versions are compatible with this straxen version."""
    cmt = straxen.CorrectionsManagementServices()
    cmt_versions = list(cmt.global_versions)[::-1]
    print(cmt_versions)
    success_for = []
    for global_version in cmt_versions:
        try:
            xenonnt(global_version)
            success_for.append(global_version)
        except straxen.CMTVersionError:
            pass
    print(
        f"This straxen version works with {success_for} but is "
        f"incompatible with {set(cmt_versions)-set(success_for)}"
    )

    test = unittest.TestCase()
    # We should always work for one offline and the online version
    test.assertTrue(len(success_for) >= 2)


##
# XENON1T
##


def test_xenon1t_dali():
    st = xenon1t_dali()
    st.search_field("time")


def test_demo():
    """Test the demo context.

    Since we download the folder to the current working directory, make sure we are in a tempfolder
    where we can write the data to

    """
    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            print("Temporary directory is ", temp_dir)
            os.chdir(temp_dir)
            st = demo()
            st.search_field("time")
        # On windows, you cannot delete the current process'
        # working directory, so we have to chdir out first.
        finally:
            os.chdir("..")


def test_fake_daq():
    st = fake_daq()
    st.search_field("time")


def test_xenon1t_led():
    st = xenon1t_led()
    st.search_field("time")



import strax
import straxen
import numpy as np
import unittest


class TestMergeIntervals(unittest.TestCase):
    def setUp(self):
        self.intervals = np.zeros(4, dtype=strax.time_fields)
        self.intervals["time"] = [2, 3, 7, 20]
        self.intervals["endtime"] = [4, 7, 8, 22]

    def test_empty_intervals(self):
        intervals = np.zeros(0, dtype=strax.time_fields)
        intervals = straxen.merge_intervals(intervals)
        assert len(intervals) == 0, "Empty input should return empty intervals!"

    def test_merge_overlapping_intervals(self):
        intervals = straxen.merge_intervals(self.intervals)

        assert len(intervals) == 2, "Got the wrong number of intervals!"

        time_is_correct = intervals[0]["time"] == self.intervals["time"][0]
        assert time_is_correct, "First interval has the wrong time!"
        time_is_correct = intervals[0]["endtime"] == self.intervals["endtime"][-2]
        assert time_is_correct, "First interval has the wrong endtime!"

        time_is_correct = intervals[-1]["time"] == self.intervals["time"][-1]
        assert time_is_correct, "Second interval has the wrong time!"
        time_is_correct = intervals[-1]["endtime"] == self.intervals["endtime"][-1]
        assert time_is_correct, "Second interval has the wrong endtime!"


class TestCoincidence(unittest.TestCase):
    def setUp(self):
        self.intervals = np.zeros(8, dtype=strax.time_fields)
        self.intervals["time"] = [3, 6, 9, 12, 15, 18, 21, 38]
        self.intervals["endtime"] = [5, 8, 10, 13, 16, 19, 23, 42]

    def test_empty_inputs(self):
        intervals = np.zeros(0, dtype=strax.time_fields)
        intervals = straxen.find_coincidence(intervals)
        assert len(intervals) == 0, "Empty input should return empty intervals!"

    def test_without_coincidence(self):
        resolving_time = 10
        truth_time = np.array([self.intervals["time"][0], self.intervals["time"][-1]])
        truth_endtime = (
            np.array([self.intervals["time"][-2], self.intervals["time"][-1]]) + resolving_time
        )
        self._test_coincidence(
            resolving_time=resolving_time,
            coincidence=1,
            pre_trigger=0,
            n_concidences_truth=2,
            times_truth=truth_time,
            endtime_truth=truth_endtime,
        )
        pre_trigger = 2
        self._test_coincidence(
            resolving_time=resolving_time,
            coincidence=1,
            pre_trigger=pre_trigger,
            n_concidences_truth=2,
            times_truth=truth_time - pre_trigger,
            endtime_truth=truth_endtime,
        )

    def test_even_fold(self):
        resolving_time = 10
        self._test_coincidence(
            resolving_time=resolving_time,
            coincidence=2,
            pre_trigger=0,
            n_concidences_truth=1,
            times_truth=self.intervals["time"][0],
            endtime_truth=self.intervals["time"][-3] + resolving_time,
        )
        pre_trigger = 2
        self._test_coincidence(
            resolving_time=resolving_time,
            coincidence=2,
            pre_trigger=pre_trigger,
            n_concidences_truth=1,
            times_truth=self.intervals["time"][0] - pre_trigger,
            endtime_truth=self.intervals["time"][-3] + resolving_time,
        )

        self._test_coincidence(
            resolving_time=resolving_time,
            coincidence=4,
            pre_trigger=pre_trigger,
            n_concidences_truth=1,
            times_truth=self.intervals["time"][0] - pre_trigger,
            endtime_truth=self.intervals["time"][-5] + resolving_time,
        )

    def test_odd_fold(self):
        resolving_time = 10
        self._test_coincidence(
            resolving_time=resolving_time,
            coincidence=3,
            pre_trigger=0,
            n_concidences_truth=1,
            times_truth=self.intervals["time"][0],
            endtime_truth=self.intervals["time"][-4] + resolving_time,
        )
        pre_trigger = 2
        self._test_coincidence(
            resolving_time=resolving_time,
            coincidence=3,
            pre_trigger=pre_trigger,
            n_concidences_truth=1,
            times_truth=self.intervals["time"][0] - pre_trigger,
            endtime_truth=self.intervals["time"][-4] + resolving_time,
        )

        self._test_coincidence(
            resolving_time=resolving_time,
            coincidence=5,
            pre_trigger=pre_trigger,
            n_concidences_truth=0,
            times_truth=self.intervals["time"][:0],
            endtime_truth=self.intervals["time"][:0],
        )

    def _test_coincidence(
        self,
        resolving_time,
        coincidence,
        pre_trigger,
        n_concidences_truth,
        times_truth,
        endtime_truth,
    ):
        coincidence = straxen.find_coincidence(
            self.intervals,
            nfold=coincidence,
            resolving_time=resolving_time,
            pre_trigger=pre_trigger,
        )
        number_coincidence_correct = len(coincidence) == n_concidences_truth
        assert number_coincidence_correct, "Have not found the correct number of coincidences"

        time_is_correct = np.all(coincidence["time"] == times_truth)
        assert time_is_correct, "Coincidence does not have the correct time"

        endtime_is_correct = np.all(coincidence["endtime"] == endtime_truth)
        print(coincidence["endtime"], endtime_truth)
        assert endtime_is_correct, "Coincidence does not have the correct endtime"


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_nv_for_dummy_rr():
    """Basic test to run the nv rr for dummy raw-records."""
    st = straxen.test_utils.nt_test_context(deregister=())
    st.context_config["forbid_creation_of"] = tuple()
    st.register(straxen.test_utils.DummyRawRecords)
    st.make(straxen.test_utils.nt_test_run_id, "hitlets_nv")



from scipy.special import logsumexp

import strax
import straxen

import numpy as np
import numba


export, __all__ = strax.exporter()


@export
class BayesPeakClassification(strax.Plugin):
    """Bayes Peak classification Returns the ln probability of a each event belonging to the S1 and
    S2 class. Uses conditional probabilities and data parameterization learned from wfsim data. More
    info can be found here xenon:xenonnt:ahiguera:bayespeakclassification.

    :param peaks: peaks
    :param waveforms: peaks waveforms in PE/ns
    :param quantiles: quantiles in ns, calculate from a cumulative sum over the waveform, from zero
        to the total area with normalized cumulative sum to determine the time
    :return: the ln probability of a each peak belonging to S1 and S2 class

    """

    provides = "peak_classification_bayes"
    depends_on = "peaks"
    __version__ = "0.0.3"
    dtype = (
        strax.time_fields
        + [("ln_prob_s1", np.float32, "S1 ln probability")]
        + [("ln_prob_s2", np.float32, "S2 ln probability")]
    )

    # Descriptor configs
    bayes_config_file = straxen.URLConfig(
        default="resource://cmt://bayes_model?version=ONLINE&run_id=plugin.run_id&fmt=npy",
        help="Bayes model, conditional probabilities tables and Bayes discrete bins",
    )
    bayes_n_nodes = straxen.URLConfig(
        default=50, help="Number of attributes(features) per waveform and quantile"
    )
    n_bayes_classes = straxen.URLConfig(default=2, help="Number of label classes S1(1)/S2(2)")

    def setup(self):
        self.class_prior = np.ones(self.n_bayes_classes) / self.n_bayes_classes
        self.bins = self.bayes_config_file["bins"]
        self.cpt = self.bayes_config_file["cprob"]

    def compute(self, peaks):
        result = np.zeros(len(peaks), dtype=self.dtype)

        waveforms, quantiles = compute_wf_and_quantiles(peaks, self.bayes_n_nodes)

        ln_prob_s1, ln_prob_s2 = compute_inference(
            self.bins,
            self.bayes_n_nodes,
            self.cpt,
            self.n_bayes_classes,
            self.class_prior,
            waveforms,
            quantiles,
        )
        result["time"] = peaks["time"]
        result["endtime"] = strax.endtime(peaks)
        result["ln_prob_s1"] = ln_prob_s1
        result["ln_prob_s2"] = ln_prob_s2
        return result


def compute_wf_and_quantiles(peaks: np.ndarray, bayes_n_nodes: int):
    """Compute waveforms and quantiles for a given number of nodes(attributes)

    :param peaks:
    :param bayes_n_nodes: number of nodes or attributes
    :return: waveforms and quantiles

    """
    data = peaks["data"].copy()
    data[data < 0.0] = 0.0
    dt = peaks["dt"]
    return _compute_wf_and_quantiles(data, dt, bayes_n_nodes)


@numba.njit(cache=True)
def _compute_wf_and_quantiles(data, sample_length, bayes_n_nodes: int):
    waveforms = np.zeros((len(data), bayes_n_nodes))
    quantiles = np.zeros((len(data), bayes_n_nodes))

    num_samples = data.shape[1]
    step_size = int(num_samples / bayes_n_nodes)
    steps = np.arange(0, num_samples + 1, step_size)
    inter_points = np.linspace(0.0, 1.0 - (1.0 / bayes_n_nodes), bayes_n_nodes)
    cumsum_steps = np.zeros(bayes_n_nodes + 1, dtype=np.float64)
    frac_of_cumsum = np.zeros(num_samples + 1)
    sample_number_div_dt = np.arange(0, num_samples + 1, 1)
    for i, (samples, dt) in enumerate(zip(data, sample_length)):
        # reset buffers
        frac_of_cumsum[:] = 0
        cumsum_steps[:] = 0

        frac_of_cumsum[1:] = np.cumsum(samples)
        frac_of_cumsum[1:] = frac_of_cumsum[1:] / frac_of_cumsum[-1]

        cumsum_steps[:-1] = np.interp(inter_points, frac_of_cumsum, sample_number_div_dt * dt)
        cumsum_steps[-1] = sample_number_div_dt[-1] * dt
        quantiles[i] = cumsum_steps[1:] - cumsum_steps[:-1]

        for j in range(bayes_n_nodes):
            waveforms[i][j] = np.sum(samples[steps[j] : steps[j + 1]])
        waveforms[i] /= step_size * dt

    return waveforms, quantiles


def compute_inference(
    bins: np.ndarray,
    bayes_n_nodes: int,
    cpt: np.ndarray,
    n_bayes_classes: int,
    class_prior: np.ndarray,
    waveforms: np.ndarray,
    quantiles: np.ndarray,
):
    """Bin the waveforms and quantiles according to Bayes bins and compute inference.

    :param bins: Bayes bins
    :param bayes_n_nodes: number of nodes or attributes
    :param cpt: conditional probability tables
    :param n_bayes_classes: number of classes
    :param class_prior: class_prior
    :param waveforms: waveforms
    :param quantiles: quantiles
    :return: ln probability per class S1/S2

    """
    # Bin the waveforms and quantiles.
    waveform_bin_edges = bins[0, :][bins[0, :] > -1]
    waveform_num_bin_edges = len(waveform_bin_edges)
    quantile_bin_edges = bins[1, :][bins[1, :] > -1]
    quantile_num_bin_edges = len(quantile_bin_edges)
    waveform_values = np.digitize(waveforms, bins=waveform_bin_edges) - 1
    waveform_values = np.clip(waveform_values, 0, waveform_num_bin_edges - 2)

    quantile_values = np.digitize(quantiles, bins=quantile_bin_edges) - 1
    quantile_values = np.clip(quantile_values, 0, quantile_num_bin_edges - 2)

    lnposterior = get_log_posterior(
        bayes_n_nodes,
        waveforms,
        cpt,
        waveform_num_bin_edges,
        quantile_num_bin_edges,
        n_bayes_classes,
        waveform_values,
        quantile_values,
    )

    lnposterior_sumsamples = np.sum(lnposterior, axis=1)
    lnposterior_sumsamples = lnposterior_sumsamples + np.log(class_prior)

    # If you need to numbafy, use
    # github.com/XENONnT/straxen/blob/c06a611a5a4289c65b9830fc1e68027149b6a5fd/straxen/plugins/bayes_peak_classification.py#L119  # noqa
    normalization = logsumexp(lnposterior_sumsamples, axis=1)
    ln_prob_s1 = lnposterior_sumsamples[:, 0] - normalization
    ln_prob_s2 = lnposterior_sumsamples[:, 1] - normalization
    return ln_prob_s1, ln_prob_s2


def get_log_posterior(
    bayes_n_nodes: int,
    waveforms: np.ndarray,
    cpt: np.ndarray,
    waveform_num_bin_edges: int,
    quantile_num_bin_edges: int,
    n_bayes_classes: int,
    waveform_values: np.ndarray,
    quantile_values: np.ndarray,
) -> np.ndarray:
    """# TODO, add a description what we are computing here

    :param bayes_n_nodes: number of nodes or attributes
    :param waveforms: waveforms
    :param cpt: conditional probability tables
    :param waveform_num_bin_edges: number of bins for waveforms
    :param quantile_num_bin_edges: number of bins for quantiles
    :param n_bayes_classes: number of classes
    :param waveform_values: digitized waveforms
    :param quantile_values: digitized quantiles
    :return: log-posterior for waveforms and quantiles. NB! This is not normalized

    """
    wf_posterior = _get_log_posterior(
        nodes=bayes_n_nodes,
        n_bins=waveform_num_bin_edges - 1,
        n_classes=n_bayes_classes,
        cpt=cpt[:bayes_n_nodes],
        wf_len=len(waveforms),
        wf_values=waveform_values,
    )
    quantile_posterior = _get_log_posterior(
        nodes=bayes_n_nodes,
        n_bins=quantile_num_bin_edges - 1,
        n_classes=n_bayes_classes,
        cpt=cpt[bayes_n_nodes:],
        wf_len=len(waveforms),
        wf_values=quantile_values,
    )

    lnposterior = np.zeros((len(waveforms), bayes_n_nodes * 2, n_bayes_classes))
    lnposterior[:, :bayes_n_nodes] = wf_posterior
    lnposterior[:, bayes_n_nodes:] = quantile_posterior
    return lnposterior


@numba.njit
def _get_log_posterior(nodes, n_bins, n_classes, cpt, wf_len, wf_values):
    """Wrapper for extracting the ln posterior inside compute_inference."""
    ln_posterior = np.zeros((wf_len, nodes, n_classes))
    for i in range(nodes):
        distribution = cpt[i, :n_bins, :]
        ln_posterior[:, i, :] = np.log(distribution[wf_values[:, i], :])
    return ln_posterior



import strax
from strax.testutils import several_fake_records
import straxen
import numpy as np
from hypothesis import given, settings


@given(several_fake_records)
@settings(deadline=None)
def test_ext_timings_nv(records):
    """Little test for nVetoExtTimings."""
    if not straxen.utilix_is_configured():
        return
    # several fake records do not have any pulse length
    # and channel start at 0, convert to nv:
    records["pulse_length"] = records["length"]
    records["channel"] += 2000

    st = straxen.contexts.xenonnt_led()
    plugin = st.get_single_plugin("1", "ext_timings_nv")
    hits = strax.find_hits(records, min_amplitude=1)

    hitlets = np.zeros(len(hits), strax.hitlet_dtype())
    strax.copy_to_buffer(hits, hitlets, "_refresh_hit_to_hitlets")
    result = plugin.compute(hitlets, records)
    assert len(result) == len(hits)
    assert np.all(result["time"] == hits["time"])
    assert np.all(result["pulse_i"] == hits["record_i"])
    true_dt = hits["time"] - records[hits["record_i"]]["time"]
    assert np.all(result["delta_time"] == true_dt)



import numpy as np
import strax
import straxen

from straxen.plugins.defaults import FAKE_MERGED_S2_TYPE

export, __all__ = strax.exporter()


@export
class Peaks(strax.Plugin):
    """Merge peaklets and merged S2s such that we obtain our peaks (replacing all peaklets that were
    later re-merged as S2s).

    As this step is computationally trivial, never save this plugin.

    """

    __version__ = "0.1.2"

    depends_on = ("peaklets", "peaklet_classification", "merged_s2s")
    data_kind = "peaks"
    provides = "peaks"
    parallel = True
    compressor = "zstd"
    save_when = strax.SaveWhen.EXPLICIT

    diagnose_sorting = straxen.URLConfig(
        track=False,
        default=False,
        infer_type=False,
        help="Enable runtime checks for sorting and disjointness",
    )

    merge_without_s1 = straxen.URLConfig(
        default=True,
        infer_type=False,
        help=(
            "If true, S1s will be igored during the merging. "
            "It's now possible for a S1 to be inside a S2 post merging"
        ),
    )

    def infer_dtype(self):
        return self.deps["peaklets"].dtype_for("peaklets")

    def compute(self, peaklets, merged_s2s):
        # Remove fake merged S2s from dirty hack, see above
        merged_s2s = merged_s2s[merged_s2s["type"] != FAKE_MERGED_S2_TYPE]

        if self.merge_without_s1:
            is_s1 = peaklets["type"] == 1
            peaks = strax.replace_merged(peaklets[~is_s1], merged_s2s)
            peaks = strax.sort_by_time(np.concatenate([peaklets[is_s1], peaks]))
        else:
            peaks = strax.replace_merged(peaklets, merged_s2s)

        if self.diagnose_sorting:
            assert np.all(np.diff(peaks["time"]) >= 0), "Peaks not sorted"
            if self.merge_without_s1:
                to_check = peaks["type"] != 1
            else:
                to_check = peaks["type"] != FAKE_MERGED_S2_TYPE

            assert np.all(
                peaks["time"][to_check][1:] >= strax.endtime(peaks)[to_check][:-1]
            ), "Peaks not disjoint"

        result = np.zeros(len(peaks), self.dtype)
        strax.copy_to_buffer(peaks, result, "_copy_requested_peak_fields")
        return result



from . import peaklets
from .peaklets import *

from . import peaklet_classification
from .peaklet_classification import *

from . import peaklet_classification_som
from .peaklet_classification_som import *



from straxen.plugins.raw_records.daqreader import DAQReader
from immutabledict import immutabledict


class Fake1TDAQReader(DAQReader):
    provides = ("raw_records", "raw_records_diagnostic", "raw_records_aqmon")

    data_kind = immutabledict(zip(provides, provides))



import strax
import numpy as np

export, __all__ = strax.exporter()


@export
class nVETOEventsSync(strax.OverlapWindowPlugin):
    """Plugin which computes time stamps which are synchronized with the TPC.

    Uses delay set in the DAQ.

    """

    depends_on = ("events_nv", "detector_time_offsets")
    delay_field_name = "time_offset_nv"

    provides = "events_sync_nv"
    save_when = strax.SaveWhen.EXPLICIT
    __version__ = "0.0.3"

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_fields
        dtype += [
            (
                (
                    "Time of the event synchronized according to the total digitizer delay.",
                    "time_sync",
                ),
                np.int64,
            ),
            (
                (
                    "Endtime of the event synchronized according to the total digitizer delay.",
                    "endtime_sync",
                ),
                np.int64,
            ),
        ]
        return dtype

    def get_window_size(self):
        # Ensure to have at least 12 offset-values from detector_time_offsets
        # to compute average time delay. Otherwise we may get unlucky with
        # our pacemaker (unlikely but could happen).
        return 120 * 10**9

    def compute(self, events_nv, detector_time_offsets):
        delay = detector_time_offsets[self.delay_field_name]
        delay = np.median(delay[delay > 0])
        delay = delay.astype(np.int64)
        # Check if delay is >= 0 otherwise something went wrong with
        # the sync signal.
        assert delay >= 0, f"Missing the GPS sync signal for run {self.run_id}."

        events_sync_nv = np.zeros(len(events_nv), self.dtype)
        events_sync_nv["time"] = events_nv["time"]
        events_sync_nv["endtime"] = events_nv["endtime"]
        events_sync_nv["time_sync"] = events_nv["time"] + delay
        events_sync_nv["endtime_sync"] = events_nv["endtime"] + delay
        return events_sync_nv



import strax
import numpy as np
from straxen.plugins.peaks.peaks import Peaks

export, __all__ = strax.exporter()


@export
class PeaksSOM(Peaks):
    """Same as Peaks but include in addition SOM type field to be propagated to event_basics.

    Thus, only change dtype.

    """

    __version__ = "0.0.1"
    child_plugin = True

    def infer_dtype(self):
        peaklet_classification_dtype = self.deps["peaklet_classification"].dtype_for(
            "peaklet_classification"
        )
        peaklets_dtype = self.deps["peaklets"].dtype_for("peaklets")
        # The merged dtype is argument position dependent!
        # It must be first classification then peaklet
        # Otherwise strax will raise an error when checking for the returned dtype!
        merged_s2s_dtype = strax.merged_dtype((peaklet_classification_dtype, peaklets_dtype))
        return merged_s2s_dtype

    def compute(self, peaklets, merged_s2s):
        result = super().compute(peaklets, merged_s2s)

        # For merged S2s SOM and straxen type are undefined:
        _is_merged_s2 = np.isin(result["time"], merged_s2s["time"]) & np.isin(
            strax.endtime(result), strax.endtime(merged_s2s)
        )
        result["straxen_type"][_is_merged_s2] = -1
        result["som_sub_type"][_is_merged_s2] = -1

        return result


@export
class PeaksSOMClassification(strax.Plugin):
    """Plugin which propagates S1 SOM infromation to peaks if straxen.PeakClassification is still
    used."""

    depends_on = ("peak_basics", "peaklet_classification_som")
    __version__ = "0.0.1"

    provides = "peak_som_classification"

    def infer_dtype(self):
        dtype_peaklets = strax.time_fields + [
            ("som_sub_type", np.int32, "SOM subtype of the peak(let)"),
            ("som_type", np.int8, "SOM type of the peak(let)"),
            ("loc_x_som", np.int16, "x location of the peak(let) in the SOM"),
            ("loc_y_som", np.int16, "y location of the peak(let) in the SOM"),
        ]
        return dtype_peaklets

    def compute(self, peaklets, peaks):
        result = np.zeros(len(peaks), self.dtype)
        result[:] = -1
        result["time"] = peaks["time"]
        result["endtime"] = peaks["endtime"]

        # Only select S1 and update their SOM fields:
        _is_s1_peaks = peaks["type"] == 1
        _is_s1_som_peaklets = peaklets["straxen_type"] == 1
        _are_same_s1_peaks = np.all(
            result[_is_s1_peaks]["time"] == peaklets[_is_s1_som_peaklets]["time"]
        )
        assert _are_same_s1_peaks, "S1 peaks and SOM S1 peaklets are not identical?!?"

        # Copy "manually" because only subset of data which does not work
        # with copy to buffer:
        for field in ("som_sub_type", "som_type", "loc_x_som", "loc_y_som"):
            result[field][_is_s1_peaks] = peaklets[field][_is_s1_som_peaklets]

        return result



"""Some shared defaults."""

DEFAULT_POSREC_ALGO = "mlp"

HE_PREAMBLE = """High energy channels: attenuated signals of the top PMT-array\n"""

MV_PREAMBLE = "Muno-Veto Plugin: Same as the corresponding nVETO-PLugin.\n"

NV_HIT_DEFAULTS = {
    "save_outside_hits_nv": (3, 15),
    "hit_min_amplitude_nv": "cmt://hit_thresholds_nv?version=ONLINE&run_id=plugin.run_id",
}

MV_HIT_DEFAULTS = {
    "save_outside_hits_mv": (2, 5),
    "hit_min_amplitude_mv": "cmt://hit_thresholds_mv?version=ONLINE&run_id=plugin.run_id",
}

FAKE_MERGED_S2_TYPE = -42



from immutabledict import immutabledict
import strax
import straxen
from straxen.common import pax_file
from .plugins_1t.fake_daqreader import Fake1TDAQReader


def get_x1t_context_config():
    """Wrapper to get straxen.contexts after imports."""
    from straxen.contexts import common_opts

    x1t_context_config = {
        **common_opts,
        **dict(
            check_available=("raw_records", "records", "peaklets", "events", "event_info"),
            free_options=("channel_map",),
            use_per_run_defaults=True,
            store_run_fields=tuple(
                [x for x in common_opts["store_run_fields"] if x != "mode"]
                + ["trigger.events_built", "reader.ini.name"]
            ),
        ),
    }
    x1t_context_config.update(
        dict(
            register=common_opts["register"]
            + [
                straxen.PeakPositions1T,
                straxen.RecordsFromPax,
                straxen.EventInfo1T,
            ]
        )
    )
    return x1t_context_config


x1t_common_config = dict(
    check_raw_record_overlaps=False,
    allow_sloppy_chunking=True,
    n_tpc_pmts=248,
    n_top_pmts=127,
    channel_map=immutabledict(
        # (Minimum channel, maximum channel)
        tpc=(0, 247),
        diagnostic=(248, 253),
        aqmon=(254, 999),
    ),
    # Records
    hev_gain_model="cmt://to_pe_model?version=v1&detector=1t&run_id=plugin.run_id",
    pmt_pulse_filter=(
        0.012,
        -0.119,
        2.435,
        -1.271,
        0.357,
        -0.174,
        -0.0,
        -0.036,
        -0.028,
        -0.019,
        -0.025,
        -0.013,
        -0.03,
        -0.039,
        -0.005,
        -0.019,
        -0.012,
        -0.015,
        -0.029,
        0.024,
        -0.007,
        0.007,
        -0.001,
        0.005,
        -0.002,
        0.004,
        -0.002,
    ),
    hit_min_amplitude="legacy-thresholds://XENON1T_SR1",
    tail_veto_threshold=int(1e5),
    save_outside_hits=(3, 3),
    # Peaklets
    peaklet_gap_threshold=350,
    gain_model="cmt://to_pe_model?version=v1&detector=1t&run_id=plugin.run_id",
    peak_split_gof_threshold=(None, ((0.5, 1), (3.5, 0.25)), ((2, 1), (4.5, 0.4))),  # Reserved
    peak_min_pmts=2,
    # MergedS2s
    s2_merge_gap_thresholds=((1.7, 5.0e3), (4.0, 500.0), (5.0, 0.0)),
    # Peaks
    # Smaller right extension since we applied the filter
    peak_right_extension=30,
    s1_max_rise_time_post100=150,
    s1_min_coincidence=3,
    event_s1_min_coincidence=3,
    # Events*
    left_event_extension=int(0.3e6),
    right_event_extension=int(1e6),
    elife=1e6,
    electron_drift_velocity=1.3325e-4,
    max_drift_length=96.9,
    electron_drift_time_gate=1700,
    se_gain=28.2,
    avg_se_gain=28.2,
    rel_extraction_eff=1.0,
    rel_light_yield=1.0,
    s1_xyz_map=f'itp_map://resource://{pax_file("XENON1T_s1_xyz_lce_true_kr83m_SR1_pax-680_fdc-3d_v0.json")}?fmt=json',  # noqa
    s2_xy_map=f'itp_map://resource://{pax_file("XENON1T_s2_xy_ly_SR1_v2.2.json")}?fmt=json',
    g1=0.1426,
    g2=11.55 / (1 - 0.63),
)


def demo():
    """Return strax context used in the straxen demo notebook."""
    straxen.download_test_data()

    st = strax.Context(
        storage=[
            strax.DataDirectory("./strax_data"),
            strax.DataDirectory(
                "./strax_test_data", deep_scan=True, provide_run_metadata=True, readonly=True
            ),
        ],
        forbid_creation_of=straxen.daqreader.DAQReader.provides,
        config=dict(**x1t_common_config),
        **get_x1t_context_config(),
    )

    # Use configs that are always available
    st.set_config(
        dict(
            hev_gain_model="legacy-to-pe://1T_to_pe_placeholder",
            gain_model="legacy-to-pe://1T_to_pe_placeholder",
            elife=1e6,
            electron_drift_velocity=1.3325e-4,
            se_gain=28.2,
            avg_se_gain=28.2,
            rel_extraction_eff=1.0,
            s1_xyz_map=(
                "itp_map://resource://"
                f"{pax_file('XENON1T_s1_xyz_lce_true_kr83m_SR1_pax-680_fdc-3d_v0.json')}"
                "?fmt=json"
            ),
            s2_xy_map=f'itp_map://resource://{pax_file("XENON1T_s2_xy_ly_SR1_v2.2.json")}?fmt=json',
        )
    )
    return st


def fake_daq():
    """Context for processing fake DAQ data in the current directory."""
    st = strax.Context(
        storage=[
            strax.DataDirectory("./strax_data"),
            # Fake DAQ puts run doc JSON in same folder:
            strax.DataDirectory("./from_fake_daq", provide_run_metadata=True, readonly=True),
        ],
        config=dict(
            daq_input_dir="./from_fake_daq",
            daq_chunk_duration=int(2e9),
            daq_compressor="lz4",
            n_readout_threads=8,
            daq_overlap_chunk_duration=int(2e8),
            **x1t_common_config,
        ),
        **get_x1t_context_config(),
    )
    st.register(Fake1TDAQReader)
    return st


def xenon1t_dali(output_folder="./strax_data", build_lowlevel=False, **kwargs):
    context_options = {**get_x1t_context_config(), **kwargs}

    st = strax.Context(
        storage=[
            strax.DataDirectory(
                "/dali/lgrandi/xenon1t/strax_converted/raw",
                take_only="raw_records",
                provide_run_metadata=True,
                readonly=True,
            ),
            strax.DataDirectory("/dali/lgrandi/xenon1t/strax_converted/processed", readonly=True),
            strax.DataDirectory(output_folder),
        ],
        config=dict(**x1t_common_config),
        # When asking for runs that don't exist, throw an error rather than
        # starting the pax converter
        forbid_creation_of=(
            straxen.daqreader.DAQReader.provides
            if build_lowlevel
            else straxen.daqreader.DAQReader.provides + ("records", "peaklets")
        ),
        **context_options,
    )
    return st


def xenon1t_led(**kwargs):
    st = xenon1t_dali(**kwargs)
    st.set_context_config({
        "check_available": ("raw_records", "led_calibration"),
        "free_options": list(get_x1t_context_config().keys()),
    })
    # Return a new context with only raw_records and led_calibration registered
    st = st.new_context(replace=True, config=st.config, storage=st.storage, **st.context_config)
    st.register([straxen.RecordsFromPax, straxen.LEDCalibration])
    return st



import strax

export, __all__ = strax.exporter()


@export
class EventInfo1T(strax.MergeOnlyPlugin):
    """Plugin which merges the information of all event data_kinds into a single data_type.

    This only uses 1T data-types as several event-plugins are nT only

    """

    depends_on = (
        "event_basics",
        "event_positions",
        "corrected_areas",
        "energy_estimates",
    )
    provides = "event_info"
    save_when = strax.SaveWhen.ALWAYS
    __version__ = "0.0.1"



import pytz
import numpy as np
import straxen
import unittest
import requests


class SCInterfaceTest(unittest.TestCase):
    def setUp(self):
        self.resources_available()
        # Simple query test:
        # Query 5 s of data:
        self.start = 1609682275000000000
        # Add micro-second to check if query does not fail if inquery precsion > SC precision
        self.start += 10**6
        self.end = self.start + 5 * 10**9

    def test_wrong_querries(self):
        parameters = {"SomeParameter": "XE1T.CTPC.Board06.Chan011.VMon"}

        with self.assertRaises(ValueError):
            # Runid but no context
            self.sc.get_scada_values(
                parameters,
                run_id="1",
                every_nth_value=1,
                query_type_lab=False,
            )

        with self.assertRaises(ValueError):
            # No time range specified
            self.sc.get_scada_values(
                parameters,
                every_nth_value=1,
                query_type_lab=False,
            )

        with self.assertRaises(ValueError):
            # Start larger end
            self.sc.get_scada_values(
                parameters,
                start=2,
                end=1,
                every_nth_value=1,
                query_type_lab=False,
            )

        with self.assertRaises(ValueError):
            # Start and/or end not in ns unix time
            self.sc.get_scada_values(
                parameters,
                start=1,
                end=2,
                every_nth_value=1,
                query_type_lab=False,
            )

    def test_pmt_names(self):
        """Tests different query options for pmt list."""
        pmts_dict = self.sc.find_pmt_names(pmts=12, current=True)
        assert "PMT12_HV" in pmts_dict.keys()
        assert "PMT12_I" in pmts_dict.keys()
        assert pmts_dict["PMT12_HV"] == "XE1T.CTPC.BOARD04.CHAN003.VMON"

        pmts_dict = self.sc.find_pmt_names(pmts=(12, 13))
        assert "PMT12_HV" in pmts_dict.keys()
        assert "PMT13_HV" in pmts_dict.keys()

        with self.assertRaises(ValueError):
            self.sc.find_pmt_names(pmts=12, current=False, hv=False)

    def test_token_expires(self):
        self.sc.token_expires_in()

    def test_convert_timezone(self):
        parameters = {"SomeParameter": "XE1T.CTPC.Board06.Chan011.VMon"}
        df = self.sc.get_scada_values(
            parameters,
            start=self.start,
            end=self.end,
            every_nth_value=1,
            query_type_lab=False,
        )

        df_strax = straxen.convert_time_zone(df, tz="strax")
        assert df_strax.index.dtype.type is np.int64

        df_etc = straxen.convert_time_zone(df, tz="Etc/GMT+0")
        assert df_etc.index.dtype.tz is pytz.timezone("Etc/GMT+0")

    def test_query_sc_values(self):
        """Unity test for the SCADAInterface.

        Query a fixed range and check if return is correct.

        """
        print("Testing SCADAInterface")
        parameters = {"SomeParameter": "XE1T.CTPC.Board06.Chan011.VMon"}
        df = self.sc.get_scada_values(
            parameters,
            start=self.start,
            end=self.end,
            every_nth_value=1,
            query_type_lab=False,
        )

        assert df["SomeParameter"][0] // 1 == 1253, "First values returned is not corrrect."
        assert np.all(np.isnan(df["SomeParameter"][1:])), "Subsequent values are not correct."

        print("Testing forwardfill option:")
        parameters = {"SomeParameter": "XE1T.CRY_FCV104FMON.PI"}
        df = self.sc.get_scada_values(
            parameters,
            start=self.start,
            end=self.end,
            fill_gaps="forwardfill",
            every_nth_value=1,
            query_type_lab=False,
        )
        assert np.all(
            np.isclose(df[:4], 2.079859)
        ), "First four values deviate from queried values."
        assert np.all(np.isclose(df[4:], 2.117820)), "Last two values deviate from queried values."
        print("Testing interpolation option:")
        self.sc.get_scada_values(
            parameters,
            start=self.start,
            end=self.end,
            fill_gaps="interpolation",
            every_nth_value=1,
            query_type_lab=False,
        )

        print("Testing down sampling and averaging option:")
        parameters = {"SomeParameter": "XE1T.CRY_TE101_TCRYOBOTT_AI.PI"}
        df_all = self.sc.get_scada_values(
            parameters,
            start=self.start,
            end=self.end,
            fill_gaps="forwardfill",
            every_nth_value=1,
            query_type_lab=False,
        )

        df = self.sc.get_scada_values(
            parameters,
            start=self.start,
            end=self.end,
            fill_gaps="forwardfill",
            down_sampling=True,
            every_nth_value=2,
            query_type_lab=False,
        )

        assert np.all(df_all[::2] == df), "Downsampling did not return the correct values."

        df = self.sc.get_scada_values(
            parameters,
            start=self.start,
            end=self.end,
            fill_gaps="forwardfill",
            every_nth_value=2,
            query_type_lab=False,
        )

        # Compare average for each second value:
        for ind, i in enumerate([0, 2, 4]):
            is_correct = np.isclose(np.mean(df_all[i : i + 2]), df["SomeParameter"][ind])
            assert is_correct, "Averaging is incorrect."

        # Testing lab query type:
        df = self.sc.get_scada_values(
            parameters,
            start=self.start,
            end=self.end,
            query_type_lab=True,
        )
        is_sorrect = np.all(df["SomeParameter"] // 1 == -96)
        assert is_sorrect, "Not all values are correct for query type lab."

    def test_long_query(self):
        """Test which querries a bit more than 35000 values over a longer period.

        Checks if additional values are querried correctly.
        See also: xenon:xenon1t:slowcontrol:webservicenew

        """
        st = straxen.contexts.xenonnt_online()
        self.sc.context = st
        res = self.sc.get_scada_values(
            parameters={"test": "XE1T.CRY_PT101_PCHAMBER_AI.PI"},
            run_id=["020000", "035000"],
            every_nth_value=500,
        )
        assert np.sum(res.test >= 0) > 35000

    @staticmethod
    def test_average_scada():
        t = np.arange(0, 100, 10)
        t_t, t_a = straxen.scada._average_scada(t / 1e9, t, 1)
        assert len(t_a) == len(t), "Scada deleted some of my 10 datapoints!"

    def resources_available(self):
        """Exception to skip test if external requirements are not met.

        Otherwise define Scada interface as self.sc.

        """
        if not straxen.utilix_is_configured(
            "scada",
            "scdata_url",
        ):
            self.skipTest("Cannot test scada since we have no access to xenon secrets.)")

        try:
            self.sc = straxen.SCADAInterface(use_progress_bar=False)
            self.sc.get_new_token()
        except requests.exceptions.SSLError:
            self.skipTest("Cannot reach database since HTTPs certifcate expired.")



from straxen import _is_on_pytest


def test_testing_suite():
    """Make sure that we are always on a pytest when this is called. E.g. pytest
    tests/test_testing_suite.py.

    A counter example would e.g. be:     python tests/test_testing_suite.py

    """
    assert _is_on_pytest()


if __name__ == "__main__":
    test_testing_suite()



import setuptools


def open_requirements(path):
    with open(path) as f:
        requires = [r.split("/")[-1] if r.startswith("git+") else r for r in f.read().splitlines()]
    return requires


# Get requirements from requirements.txt, stripping the version tags
requires = open_requirements("requirements.txt")
tests_requires = open_requirements("extra_requirements/requirements-tests.txt")
doc_requirements = open_requirements("extra_requirements/requirements-docs.txt")

with open("README.md") as file:
    readme = file.read()

with open("HISTORY.md") as file:
    history = file.read()

setuptools.setup(
    name="straxen",
    version="2.2.0",
    description="Streaming analysis for XENON",
    author="Straxen contributors, the XENON collaboration",
    url="https://github.com/XENONnT/straxen",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/markdown",
    setup_requires=["pytest-runner"],
    install_requires=requires,
    tests_require=requires + tests_requires,
    python_requires=">=3.9",
    extras_require={
        "docs": doc_requirements,
        "microstrax": ["hug"],
    },
    scripts=[
        "bin/bootstrax",
        "bin/restrax",
        "bin/straxer",
        "bin/fake_daq",
        "bin/microstrax",
        "bin/ajax",
        "bin/refresh_raw_records",
        "bin/straxen-print_versions",
    ],
    packages=setuptools.find_packages() + ["extra_requirements"],
    package_dir={"extra_requirements": "extra_requirements"},
    package_data={"extra_requirements": ["requirements-docs.txt", "requirements-tests.txt"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    zip_safe=False,
)



"""Testing functions for the CMT services."""

import strax
import straxen
import utilix
import numpy as np
from .test_basics import test_run_id_1T
from straxen.test_utils import nt_test_run_id as test_run_id_nT
from straxen.common import aux_repo
import unittest


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_connect_to_db():
    """Test connection to db."""
    corrections_collection = utilix.rundb.xent_collection(database="corrections")
    client = corrections_collection.database.client
    cmt = strax.CorrectionsInterface(client, database_name="corrections")
    df = cmt.read("global_xenonnt")
    mes = "Return empty dataframe when reading DB. Please check"
    assert not df.empty, mes


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_1T_elife():
    """Test elife from CMT DB against historical data(aux file)"""
    elife_conf = ("elife_xenon1t", "ONLINE", False)
    elife_cmt = straxen.get_correction_from_cmt(test_run_id_1T, elife_conf)
    elife_file = aux_repo + "3548132b55f81a43654dba5141366041e1daaf01/strax_files/elife.npy"
    x = straxen.get_resource(elife_file, fmt="npy")
    run_index = np.where(x["run_id"] == int(test_run_id_1T))[0]
    elife = x[run_index[0]]["e_life"]
    mes = "Elife values do not match. Please check"
    assert elife_cmt == elife, mes


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_cmt_conf_option(option="mlp_model", version="ONLINE", is_nT=True):
    """Test CMT conf options If wrong conf is passed it would raise an error accordingly."""
    conf = option, version, is_nT
    correction = straxen.get_correction_from_cmt(test_run_id_nT, conf)
    assert isinstance(correction, (float, int, str, np.ndarray))


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_mc_wrapper_elife(
    run_id="009000",
    cmt_id="016000",
    mc_id="mc_0",
):
    """Test that for two different run ids, we get different elifes using the MC wrapper.

    :param run_id: First run-id (used for normal query)
    :param cmt_id: Second run-id used as a CMT id (should not be the same as run_id! otherwise the
        values might actually be the same and the test does not work).
    :return: None

    """
    assert np.abs(int(run_id) - int(cmt_id)) > 500, "runs must be far apart"

    # First for the run-id let's get the value
    elife = straxen.get_correction_from_cmt(run_id, ("elife", "ONLINE", True))

    # Now, we repeat the same query using the MC wrapper, this should
    # give us a different result since we are now asking for a very
    # different run-number.
    mc_elife_diff = straxen.get_correction_from_cmt(
        mc_id, ("cmt_run_id", cmt_id, "elife", "ONLINE", True)
    )

    # Repeat the query from above to verify, let's see if we are getting
    # the same results as for `elife` above
    mc_elife_same = straxen.get_correction_from_cmt(
        mc_id, ("cmt_run_id", run_id, "elife", "ONLINE", True)
    )

    assert elife != mc_elife_diff
    assert elife == mc_elife_same


@unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
def test_mc_wrapper_gains(
    run_id="009000",
    cmt_id="016000",
    mc_id="mc_0",
    execute=True,
):
    """Test that for two different run ids, we get different gains using the MC wrapper.

    :param run_id: First run-id (used for normal query)
    :param cmt_id: Second run-id used as a CMT id (should not be the same as run_id! otherwise the
        values might actually be the same and the test does not work).
    :param execute: Execute this test (this is set to False since the test takes 9 minutes which is
        too long. We can activate this if the testing time due to faster CMT queries is reduced).
    :return: None

    """
    if not execute:
        return
    assert np.abs(int(run_id) - int(cmt_id)) > 500, "runs must be far apart"

    # First for the run-id let's get the value
    gains = straxen.get_correction_from_cmt(run_id, ("to_pe_model", "ONLINE", True))

    # Now, we repeat the same query using the MC wrapper, this should
    # give us a different result since we are now asking for a very
    # different run-number.
    mc_gains_diff = straxen.get_correction_from_cmt(
        mc_id, ("cmt_run_id", cmt_id, "to_pe_model", "ONLINE", True)
    )

    # Repeat the query from above to verify, let's see if we are getting
    # the same results as for `gains` above
    mc_gains_same = straxen.get_correction_from_cmt(
        mc_id, ("cmt_run_id", run_id, "to_pe_model", "ONLINE", True)
    )

    assert not np.all(gains == mc_gains_diff)
    assert np.all(gains == mc_gains_same)


def test_is_cmt_option():
    """Catches if we change the CMT option structure.

    The example dummy_option works at least before Jun 13 2021

    """
    dummy_option = ("hit_thresholds_tpc", "ONLINE", True)
    assert straxen.is_cmt_option(dummy_option), "Structure of CMT options changed!"

    dummy_url_config = "cmt://correction?version=ONLINE&run_id=plugin.run_id"
    assert straxen.is_cmt_option(dummy_url_config), "Structure of CMT options changed!"


def test_replace_url_version():
    """Tests the replace_url_version function which is important in apply_cmt_version."""
    url = "cmt://elife?version=ONLINE&run_id=plugin.run_id"
    url_check = "cmt://elife?version=v1&run_id=plugin.run_id"
    url_test = straxen.corrections_services.replace_url_version(url, "v1")
    if url_check != url_test:
        raise AssertionError(
            "replace_url_version did not do its job! "
            f"it returns:\n{url_test}\nwhen it should return:\n{url_check}"
        )



from straxen.plugins.events_nv import nVETOEventsSync

import strax

export, __all__ = strax.exporter()


@export
class mVETOEventSync(nVETOEventsSync):
    """Plugin which computes synchronized timestamps for the muon-veto with respect to the TPC."""

    depends_on = ("events_mv", "detector_time_offsets")
    delay_field_name = "time_offset_mv"

    provides = "events_sync_mv"
    __version__ = "0.0.1"
    child_plugin = True

    def compute(self, events_mv, detector_time_offsets):
        return super().compute(events_mv, detector_time_offsets)



from immutabledict import immutabledict
import warnings
import panel as pn
import numpy as np
import holoviews as hv
import strax
from .holoviews_inspector import EventStats
from .holoviews_peak_data import PlotPeaksTPC
from .holoviews_pmt_array import PlotPMTArrayTPC


class InteractiveTPCEventDisplay:
    """Interactive event display for the TPC. Can be used for waveform watching/inspection. For
    waveform inspection the user can provide a subset of peak data which only belongs to the peaks
    of interest. In this case the inspector plot plots all events, but only the selected events with
    peaks can be choosen with the event slider.

    :param events:
    :param peaks:
    :param selection:
    :param gains: Can be used to apply to_pe value to area_per_channel before plotting. PMTs with a
        to_pe value of zero are displayed in gray.

    """

    _current_event_index = -1

    def __init__(
        self,
        events,
        peaks,
        gains=None,
        plot_alt_peaks=True,
    ):
        self._events = events
        self._peaks = peaks
        self._plot_alt_peaks = plot_alt_peaks
        self._split_peaks_by_event()
        self._make_event_slider()
        self.gains = gains

    def event_display(
        self,
        inspector_config=immutabledict(),
        exclude_fields=None,
        inspection_mode=True,
        plot_alt=True,
        log_hit_pattern=True,
        row_height=300,
        row_width=900,
    ):
        """Function which plots interactive event display."""

        plots = self._get_display_componnts(
            inspector_config=inspector_config,
            exclude_fields=exclude_fields,
            inspection_mode=inspection_mode,
            plot_alt=plot_alt,
            log_hit_pattern=log_hit_pattern,
        )
        pattern, inspector_plot, inspector_widget, s2_peak, s1_peak, event_display = plots

        event_display.opts(height=row_height, width=row_width)
        row_setting = dict(height=row_height, width=row_width // 3)
        s1_s2_pattern = pn.Row(
            s1_peak.opts(**row_setting), s2_peak.opts(**row_setting), pattern.opts(**row_setting)
        )
        if inspection_mode:
            inspector_widget.width = row_width // 2
            inspector_widget.height = row_height
            inspector_row = pn.Row(
                inspector_plot.opts(height=row_height, width=row_width // 2),
                pn.Column(self.event_slider, inspector_widget),
            )
            return pn.Column(s1_s2_pattern, event_display, inspector_row)
        return pn.Column(self.event_slider, s1_s2_pattern, event_display)

    def _get_display_componnts(
        self,
        inspector_config=immutabledict(),
        exclude_fields=None,
        inspection_mode=True,
        plot_alt=True,
        log_hit_pattern=True,
    ):
        """Function which creates all required plot and widget components for the interactive event
        display."""
        inspector_plot = None
        inspector_widget = None
        if inspection_mode:
            inspector_plot, inspector_widget = self._get_inspector_plot_and_widget(
                inspector_config, exclude_fields
            )

        streams = {"event_index": self.event_slider}
        pattern = hv.DynamicMap(self._pmt_pattern_callback, streams=streams)

        s2_peak = hv.DynamicMap(self._s2_callback, streams=streams)
        s1_peak = hv.DynamicMap(self._s1_callback, streams=streams)
        event_display = hv.DynamicMap(self._event_callback, streams=streams)

        return pattern, inspector_plot, inspector_widget, s2_peak, s1_peak, event_display

    def _get_inspector_plot_and_widget(self, inspector_config, exclude_fields):
        inspector = EventStats(
            self._events,
            inspector_config,
            exclude_fields,
            _index_slider=self.event_slider,
        )
        stream_hist, stream_marker = inspector._create_streams()
        hist = hv.DynamicMap(inspector._plot_hist_data, streams=stream_hist)
        selected_points = hv.DynamicMap(inspector._plot_selected_event, streams=stream_marker)

        inspector_plot = hist * selected_points
        inspector_widget = inspector._make_widget_panel(include_index_slider=False)
        return inspector_plot, inspector_widget

    def _selcet_current_event(self, event_index):
        """Function which selects current event.

        If event has been already selected do nothing. This improves performance as we need four
        different callbacks.

        """
        if self._current_event_index == event_index:
            return

        self._selected_event = self._events[event_index]
        self._peaks_in_event = self._peaks[self._fully_contained_index == event_index]

        self._s2 = self._peaks_in_event[self._selected_event["s2_index"]]
        self._s1 = self._peaks_in_event[self._selected_event["s1_index"]]
        other_peaks_index = np.arange(len(self._peaks_in_event))
        mask_other_s2 = self._peaks_in_event["type"] == 2
        mask_other_s2 &= other_peaks_index != self._selected_event["s2_index"]

        self._alt_s2 = None
        _can_plot_alt_s2 = (self._selected_event["alt_s2_index"] != -1) and self._plot_alt_peaks
        if _can_plot_alt_s2:
            self._alt_s2 = self._peaks_in_event[self._selected_event["alt_s2_index"]]
            mask_other_s2 &= other_peaks_index != self._selected_event["alt_s2_index"]

        self._alt_s1 = None
        _can_plot_alt_s1 = (self._selected_event["alt_s1_index"] != -1) and self._plot_alt_peaks
        if _can_plot_alt_s1:
            self._alt_s1 = self._peaks_in_event[self._selected_event["alt_s1_index"]]

        self._other_s2 = self._peaks_in_event[mask_other_s2]
        self._current_event_index = event_index

    def _pmt_pattern_callback(self, event_index, top_pmt_array=True, pattern_ops=immutabledict()):
        """Callback function for the PMT hitpattern.

        Selects main and alt S2 position for a given event_index and creates the plot.

        """
        self._selcet_current_event(event_index)

        pmt_array = PlotPMTArrayTPC(
            gains=self.gains,
            top_pmt_array=top_pmt_array,
        )

        s2_plot = pmt_array.plot_pmt_array(
            self._s2,
            label="S2",
            logz=True,  # TODO make me optional?
            **pattern_ops,
        )
        s2_point = hv.Points(
            (self._selected_event["x"], self._selected_event["y"]), label="S2"
        ).opts(color="red", size=5)
        position_plot = s2_plot
        position_plot = s2_plot * s2_point

        if self._alt_s2 is not None:
            alt_s2_plot = pmt_array.plot_pmt_array(
                self._alt_s2,
                label="alt. S2",
                logz=True,  # TODO make me optional?
                **pattern_ops,
            )
            alt_s2_point = hv.Points(
                (self._selected_event["alt_s2_x"], self._selected_event["alt_s2_y"]),
                label="alt. S2",
            ).opts(
                color="orange",
                size=5,
            )  # Todo allow to cutsomize better...
            # Cannot use *= because changes plot order
            position_plot = (alt_s2_plot * alt_s2_point) * position_plot

        return position_plot.opts(legend_opts={"click_policy": "hide"})

    def _peak_callback(
        self,
        main_peak,
        main_label,
        alt_peak,
        alt_label,
        time_in_µs=False,
        opts_curve=immutabledict(),
        opts_area=immutabledict(),
        amplitude_prefix="",
    ):
        """Call back which plots main/alt S1/S2 for display."""
        peak_ploter = PlotPeaksTPC()
        peak_plot = peak_ploter.plot_peak(
            main_peak,
            label=main_label,
            opts_curve=opts_curve,
            opts_area=opts_area,
            amplitude_prefix=amplitude_prefix,
            time_in_us=time_in_µs,
        )
        if alt_peak is not None:
            opts_curve = self._add_legend_mute(opts_curve)
            opts_area = self._add_legend_mute(opts_area)
            alt_peak_plot = peak_ploter.plot_peak(
                alt_peak,
                label=alt_label,
                opts_curve=opts_curve,
                opts_area=opts_area,
                amplitude_prefix=amplitude_prefix,
                time_in_us=time_in_µs,
            )

            peak_plot *= alt_peak_plot
        return peak_plot

    def _s2_callback(
        self,
        event_index,
        opts_peak_area=immutabledict(color="purple"),
        opts_peak_curve=immutabledict(color="purple"),
    ):
        """Call back which plots main/alt S2 for display."""
        self._selcet_current_event(event_index)
        return self._peak_callback(
            self._s2,
            "S2",
            self._alt_s2,
            "alt. S2",
            time_in_µs=True,
            opts_area=opts_peak_area,
            opts_curve=opts_peak_curve,
            amplitude_prefix="S2",
        ).opts(title="main/alt. S2")

    def _s1_callback(
        self,
        event_index,
        opts_peak_area=immutabledict(color="orange"),
        opts_peak_curve=immutabledict(color="orange"),
    ):
        """Call back which plots main/alt S1 for display."""
        self._selcet_current_event(event_index)
        return self._peak_callback(
            self._s1,
            "S1",
            self._alt_s1,
            "alt. S1",
            time_in_µs=False,
            opts_area=opts_peak_area,
            opts_curve=opts_peak_curve,
            amplitude_prefix="S1",
        ).opts(title="main/alt. S1")

    def _event_callback(
        self,
        event_index,
    ):
        """Callback function to plot the main display."""
        self._selcet_current_event(event_index)

        # Get event start time and extend a bit to the left:
        event_start = self._selected_event["time"]
        event_end = self._selected_event["endtime"]
        s1_start = self._selected_event["s1_time"]
        rel_start = max(event_start, s1_start)
        peak_ploter = PlotPeaksTPC()
        event_plot = []

        mask = self._peaks_in_event["type"] == 2
        self.n_s2 = np.sum(mask)
        if np.any(mask):
            plot_s2 = peak_ploter.plot_peaks(
                self._peaks_in_event[mask],
                label=f"S2 {event_index}",
                group_label=f"Event {event_index}",
                opts_curve={
                    "color": "purple",
                    "muted_alpha": 0,
                },
                opts_area={
                    "color": "purple",
                    "muted_alpha": 0,
                },
                time_in_us=True,
                time_prefix="Event",
                amplitude_prefix="Event",
                _relative_start_time=rel_start,
            )
            event_plot.append(plot_s2)

        mask = self._peaks_in_event["type"] == 1
        self.n_s1 = np.sum(mask)
        if np.any(mask):
            plot_s1 = peak_ploter.plot_peaks(
                self._peaks_in_event[mask],
                label=f"S1 {event_index}",
                group_label=f"Event {event_index}",
                opts_curve={
                    "color": "orange",
                    "muted_alpha": 0,
                },
                opts_area={
                    "color": "orange",
                    "muted_alpha": 0,
                },
                time_in_us=True,
                time_prefix="Event",
                amplitude_prefix="Event",
                _relative_start_time=rel_start,
            )
            event_plot.append(plot_s1)

        mask = self._peaks_in_event["type"] == 0
        self.n_s0 = np.sum(mask)
        if np.any(mask):
            plot_s0 = peak_ploter.plot_peaks(
                self._peaks_in_event[mask],
                label=f"S0 {event_index}",
                group_label=f"Event {event_index}",
                opts_curve={
                    "color": "gray",
                    "muted_alpha": 0,
                },
                opts_area={
                    "color": "gray",
                    "muted_alpha": 0,
                },
                time_in_us=True,
                time_prefix="Event",
                amplitude_prefix="Event",
                _relative_start_time=rel_start,
            )
            event_plot.append(plot_s0)

        shaded_regions = self._get_alt_main_shading(self._selected_event, rel_start, in_µs=True)
        event_plot = hv.Overlay(event_plot + shaded_regions)
        return event_plot.opts(
            legend_opts={
                "click_policy": "hide",
            },
            show_legend=True,
            xlim=((event_start - rel_start) / 10**3, (event_end - rel_start) / 10**3),
            ylim=(0, None),
        )

    def _get_alt_main_shading(self, event, start_time, in_µs=True):
        """Function which shades alt./main S1/S2 with."""
        time_scaler = 1
        if in_µs:
            time_scaler = 1000

        peak_times = dict()
        for peak in ["s1", "s2", "alt_s1", "alt_s2"]:
            peak_start = event["_".join((peak, "time"))]
            peak_end = event["_".join((peak, "endtime"))]
            _has_peak = peak_start != -1
            if _has_peak:
                peak_times[peak] = (
                    (peak_start - start_time) / time_scaler,
                    (peak_end - start_time) / time_scaler,
                )
            else:
                peak_times[peak] = (-start_time, -start_time)

        shaded_regions = []
        for peak_type, (start, end) in peak_times.items():
            span = hv.VSpan(start, end).opts(color="gray", alpha=0.4)
            text = hv.Text(
                start, 0, rotation=90, text=peak_type, valign="bottom", halign="left"
            ).opts(text_alpha=1)

            # In case peak does not exists make label transparent. (We have
            # to always make 4 labels as otherwise hv.Text breaks if we
            # start with less peaks.)
            _no_peak = start == -start_time
            if _no_peak:
                text.opts(text_alpha=0)

            shaded_regions += [span, text]

        return shaded_regions

    @staticmethod
    def _add_legend_mute(options):
        options = {k: v for k, v in options.items()}
        options["muted"] = True
        options["muted_alpha"] = 0
        return immutabledict(options)

    def _split_peaks_by_event(self):
        """Split provided peaks by event."""
        fci = strax.fully_contained_in(self._peaks, self._events)
        if np.any(fci == -1):
            warnings.warn(
                "Found peaks which are not belonging to any "
                "event. We are dropping these peaks for efficiency "
                "purposes."
            )
        self._peaks = self._peaks[fci != -1]
        self._fully_contained_index = fci[fci != -1]
        self._events_with_peaks = list(np.unique(self._fully_contained_index))

    def _make_event_slider(self):
        """Creates slider for all events with peak data."""
        self.event_slider = pn.widgets.DiscreteSlider(
            name="Selected Event Index:",
            options=self._events_with_peaks,
            value=self._events_with_peaks[0],
            value_throttled=True,
        )



import os
import json
import tempfile
import pandas as pd
import strax
import straxen
import fsspec
import utilix.rundb
from straxen.test_utils import nt_test_context, nt_test_run_id
import unittest
import pickle
import random
import numpy as np
from datetime import datetime


class DummyObject:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


@straxen.URLConfig.register("random")
def generate_random(_):
    return random.random()


@straxen.URLConfig.register("unpicklable")
def return_lamba(_):
    return lambda x: x


@straxen.URLConfig.register("large-array")
def large_array(_):
    return np.ones(1_000_000).tolist()


@straxen.URLConfig.register("object-list")
def object_list(length):
    length = int(length)
    return [DummyObject(a=i, b=i + 1) for i in range(length)]


@straxen.URLConfig.preprocessor
def formatter(config, **kwargs):
    if not isinstance(config, str):
        return config
    try:
        config = config.format(**kwargs)
    except KeyError:
        pass
    return config


GLOBAL_VERSIONS = {"global_v1": {"test_config": "v0"}}


@straxen.URLConfig.preprocessor
def replace_global_version(config, name=None, **kwargs):
    if name is None:
        return

    if not isinstance(config, str):
        return config

    if straxen.URLConfig.SCHEME_SEP not in config:
        return config

    version = straxen.URLConfig.kwarg_from_url(config, "version")

    if version is None:
        return config

    if version.startswith("global_") and version in GLOBAL_VERSIONS:
        version = GLOBAL_VERSIONS[version].get(name, version)
        config = straxen.URLConfig.format_url_kwargs(config, version=version)
    return config


class ExamplePlugin(strax.Plugin):
    depends_on = ()
    dtype = strax.time_fields
    provides = ("test_data",)
    test_config = straxen.URLConfig(
        default=42,
    )
    cached_config = straxen.URLConfig(default=666, cache=1)

    def compute(self):
        pass


class TestURLConfig(unittest.TestCase):
    def setUp(self):
        st = nt_test_context()
        st.register(ExamplePlugin)
        self.st = st

    def test_default(self):
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, 42)

    def test_literal(self):
        self.st.set_config({"test_config": 666})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, 666)

    def test_leading_zero_int(self):
        self.st.set_config({"test_config": "format://{value}?value=0666"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, "0666")
        self.assertIsInstance(p.test_config, str)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test CMT.")
    def test_cmt_protocol(self):
        self.st.set_config({"test_config": "cmt://elife?version=v1&run_id=plugin.run_id"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertTrue(abs(p.test_config - 219203.49884000001) < 1e-2)

    def test_json_protocol(self):
        self.st.set_config({"test_config": "json://[1,2,3]"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, [1, 2, 3])

    def test_format_protocol(self):
        self.st.set_config({"test_config": "format://{run_id}?run_id=plugin.run_id"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, nt_test_run_id)

    def test_fsspec_protocol(self):
        with fsspec.open("memory://test_file.json", mode="w") as f:
            json.dump({"value": 999}, f)
        self.st.set_config(
            {"test_config": "take://json://fsspec://memory://test_file.json?take=value"}
        )
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, 999)

    def test_chained(self):
        self.st.set_config({"test_config": "take://json://[1,2,3]?take=0"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, 1)

    def test_take_nested(self):
        self.st.set_config({"test_config": 'take://json://{"a":[1,2,3]}?take=a&take=0'})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, 1)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test!")
    def test_bodedga_get(self):
        """Just a didactic example."""
        self.st.set_config({
            "test_config": (
                "take://resource://XENONnT_numbers.json?fmt=json&take=g1&take=v2&take=value"
            )
        })
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        # Either g1 is 0, bodega changed or someone broke URLConfigs
        self.assertTrue(p.test_config)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test CMT.")
    def test_itp_dict(self, ab_value=20, cd_value=21, dump_as="json"):
        """Test that we are getting ~the same value from interpolating at the central date in a
        dict.

        :param ab_value, cd_value: some values to test against
        :param dump_as: Write as csv or as json file

        """
        central_datetime = (
            utilix.rundb.xent_collection()
            .find_one({"number": int(nt_test_run_id)}, projection={"start": 1})
            .get("start", "QUERY FAILED!")
        )
        fake_file = {
            "time": [
                datetime(2000, 1, 1).timestamp() * 1e9,
                central_datetime.timestamp() * 1e9,
                datetime(2040, 1, 1).timestamp() * 1e9,
            ],
            "ab": [10, ab_value, 30],
            "cd": [11, cd_value, 31],
        }

        temp_dir = tempfile.TemporaryDirectory()

        if dump_as == "json":
            fake_file_name = os.path.join(temp_dir.name, "test_seg.json")
            with open(fake_file_name, "w") as f:
                json.dump(fake_file, f)
        elif dump_as == "csv":
            # This example also works well with dataframes!
            fake_file_name = os.path.join(temp_dir.name, "test_seg.csv")
            pd.DataFrame(fake_file).to_csv(fake_file_name)
        else:
            raise ValueError

        self.st.set_config({
            "test_config": (
                "itp_dict://"
                "resource://"
                f"{fake_file_name}"
                "?run_id=plugin.run_id"
                f"&fmt={dump_as}"
                "&itp_keys=ab,cd"
            )
        })
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertIsInstance(p.test_config, dict)
        assert np.isclose(p.test_config["ab"], ab_value, rtol=1e-3)
        assert np.isclose(p.test_config["cd"], cd_value, rtol=1e-3)
        temp_dir.cleanup()

    def test_itp_dict_csv(self):
        self.test_itp_dict(dump_as="csv")

    def test_rekey(self):
        original_dict = {"a": 1, "b": 2, "c": 3}
        check_dict = {"anew": 1, "bnew": 2, "cnew": 3}

        temp_dir = tempfile.TemporaryDirectory()

        fake_file_name = os.path.join(temp_dir.name, "test_dict.json")
        with open(fake_file_name, "w") as f:
            json.dump(original_dict, f)

        self.st.set_config({
            "test_config": (
                f"rekey_dict://resource://{fake_file_name}?"
                "fmt=json&replace_keys=a,b,c"
                "&with_keys=anew,bnew,cnew"
            )
        })
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, check_dict)
        temp_dir.cleanup()

    def test_print_protocol_desc(self):
        straxen.URLConfig.print_protocols()

    def test_cache(self):
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")

        # sanity check that default value is not affected
        self.assertEqual(p.cached_config, 666)
        self.st.set_config({"cached_config": "random://abc"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")

        # value is randomly generated when accessed so if
        # its equal when we access it again, its coming from the cache
        cached_value = p.cached_config
        self.assertEqual(cached_value, p.cached_config)

        # now change the config to which will generate a new number
        self.st.set_config({"cached_config": "random://dfg"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")

        # sanity check that the new value is still consistent i.e. cached
        self.assertEqual(p.cached_config, p.cached_config)

        # test if previous value is evicted, since cache size is 1
        self.assertNotEqual(cached_value, p.cached_config)

        # verify pickalibility of objects in cache dont affect plugin pickalibility
        self.st.set_config({"cached_config": "unpicklable://dfg"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        with self.assertRaises(AttributeError):
            pickle.dumps(p.cached_config)
        pickle.dumps(p)

    def test_cache_size(self):
        """Test the cache helper functions."""
        # make sure the value has a detectable size
        self.st.set_config({"cached_config": "large-array://dfg"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")

        # fetch the value so its stored in the cache
        p.cached_config

        # cache should now have finite size
        self.assertGreater(straxen.config_cache_size_mb(), 0.0)

        # test if clearing cache works as expected
        straxen.clear_config_caches()
        self.assertEqual(straxen.config_cache_size_mb(), 0.0)

    def test_filter_kwargs(self):
        all_kwargs = dict(a=1, b=2, c=3)

        # test a function that takes only a seubset of the kwargs
        def func1(a=None, b=None):
            return

        filtered1 = straxen.filter_kwargs(func1, all_kwargs)
        self.assertEqual(filtered1, dict(a=1, b=2))
        func1(**filtered1)

        # test function that accepts wildcard kwargs
        def func2(**kwargs):
            return

        filtered2 = straxen.filter_kwargs(func2, all_kwargs)
        self.assertEqual(filtered2, all_kwargs)
        func2(**filtered2)

    @unittest.skipIf(not straxen.utilix_is_configured(), "No db access, cannot test CMT.")
    def test_dry_evaluation(self):
        """Check that running a dry evaluation can be done outside of the context of a URL config
        and yield the same result."""
        plugin_url = "cmt://electron_drift_velocity?run_id=plugin.run_id&version=v3"
        self.st.set_config({"test_config": plugin_url})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        correct_val = p.test_config

        # We can also get it from one of these methods
        dry_val1 = straxen.URLConfig.evaluate_dry(
            f"cmt://electron_drift_velocity?run_id={nt_test_run_id}&version=v3"
        )
        dry_val2 = straxen.URLConfig.evaluate_dry(
            f"cmt://electron_drift_velocity?version=v3", run_id=nt_test_run_id
        )

        # All methods should yield the same
        assert correct_val == dry_val1 == dry_val2

        # However dry-evaluation does NOT allow loading the plugin.run_id
        # as in the plugin_url and should complain about that
        with self.assertRaises(ValueError):
            straxen.URLConfig.evaluate_dry(plugin_url)

    def test_objects_to_dict(self):
        n = 3
        self.st.set_config(
            {"test_config": f"objects-to-dict://object-list://{n}?key_attr=a&value_attr=b"}
        )
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, {i: i + 1 for i in range(n)})

    def test_list_to_array(self):
        n = 3
        self.st.set_config({"test_config": f"list-to-array://object-list://{n}"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertIsInstance(p.test_config, np.ndarray)

    def test_format_preprocessor(self):
        self.st.set_config({"test_config": "{name}:{run_id}"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, f"test_config:{nt_test_run_id}")
        self.assertEqual(p.test_config, p.config["test_config"])

    def test_global_version_preprocessor(self):
        self.st.set_config({"test_config": "fake://url?version=global_v1"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, "fake://url?version=v0")

    def test_alphabetize_url_kwargs(self):
        """URLConfig preprocessor to rearange the order of arguments given buy a url to ensure the
        same url with a different hash order gives the same hash."""
        url = "xedocs://electron_lifetimes?run_id=034678&version=v5&attr=value"
        intended_url = "xedocs://electron_lifetimes?attr=value&run_id=034678&version=v5"
        preprocessed_url = straxen.url_config.alphabetize_url_kwargs(url)
        self.assertEqual(intended_url, preprocessed_url)

    def test_xedocs_global_version_hash_coinsistency(self):
        # Same URLs but the queries are in a different order
        st1 = self.st.new_context(
            config={"test_config": "fake://electron_lifetimes?run_id=25000&version=v5&attr=value"}
        )
        st2 = self.st.new_context(
            config={"test_config": "fake://electron_lifetimes?attr=value&run_id=25000&version=v5"}
        )
        self.assertEqual(
            st1.key_for(25000, "corrected_areas").lineage_hash,
            st2.key_for(25000, "corrected_areas").lineage_hash,
        )

    def test_global_version_not_changed(self):
        """
        - if no global version is matched, the url version should not be changed
        - if config is not matched, the url version should not be changed
        """
        assert "global_v2" not in GLOBAL_VERSIONS
        self.st.set_config({"test_config": "fake://url?version=global_v2"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, "fake://url?version=global_v2")

        class ExamplePluginNew(ExamplePlugin):
            test_config_new = straxen.URLConfig(
                default="rootisthesourceofallevil",
            )

        self.st.register(ExamplePluginNew)
        self.st.set_config({"test_config_new": "fake://url?version=global_v1"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config_new, "fake://url?version=global_v1")

    @unittest.skipIf(
        not straxen.utilix_is_configured(), "No db access, cannot test run_doc protocol."
    )
    def test_run_doc_protocol(self):
        self.st.set_config({"test_config": "run_doc://mode?run_id=plugin.run_id"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, "tpc_commissioning")

        self.st.set_config({"test_config": "run_doc://fake_key?run_id=plugin.run_id&default=42"})
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(p.test_config, 42)

        with self.assertRaises(ValueError):
            self.st.set_config({"test_config": "run_doc://mode?run_id=plugin.run_id"})
            p = self.st.get_single_plugin(999999999, "test_data")
            return p.test_config

    def test_pad_array(self):
        """Test that pad_array works as expected."""

        self.st.set_config(
            {"test_config": "pad-array://json://[1,2,3]?pad_left=2&pad_right=3&pad_value=0"}
        )
        p = self.st.get_single_plugin(nt_test_run_id, "test_data")
        self.assertEqual(len(p.test_config), 8)
        self.assertEqual(p.test_config[0], 0)
        self.assertEqual(p.test_config[-1], 0)
